import itertools
import logging
from typing import Any, Callable, Dict, List

from torch.utils.data import Dataset

from pele.config import DataConfig
from pele.types import Input

from pele.core.data.features import DataFeature


class ExperimentDataset(Dataset):
    """
    Create an experiment dataset from the data configuration. This class takes a set of features, and links them into a single PyTorch dataset.
        
    :param config: The data configuration for the experiment.
    :type config: DataConfig
    """

    def __init__(self, config: DataConfig) -> None:

        # Step 1 - Load the feature yaml files
        self.features: Dict[str, DataFeature] = {}
        for elem in config.features:
            self.features[elem['output_key']] = elem['type'](
                **{k: v for k, v in elem.items() if k != 'type'})
        self.keys = list(set(itertools.chain.from_iterable(v.keys()
                                                           for v in self.features.values())))
        if config.db_limit is not None:
            self.keys = self.keys[:config.db_limit]

    def __len__(self,) -> int:
        return len(self.keys)

    def __getitem__(self, index: int) -> Input:
        key = self.keys[index]
        output: Input = {}
        for feature_key, feature in self.features.items():
            try:
                output.update(feature[key])
            except KeyError:
                logging.warning(
                    'Dataset ignoring feature key %s for sample with key %s', feature_key, key)
        return output

    @classmethod
    def build_collate_fn(cls, train_data_config: DataConfig, val_data_config: DataConfig) -> Callable[[List[Any]], Input]:
        """
        Get a collate function for the dataset. By using training and validation methods, this function obtains a collate
        function for the entire dataset by linking the collate functions for the individual features. This must be a
        class method, since the function has to be instantiated outside the experiment runner for ray to properly
        distribute the experiment.
        
        :param train_data_config: A configuration specifying the training data features
        :type train_data_config: DataConfig
        :param val_data_config: A configuration specifying the validation data features
        :type val_data_config: DataConfig
        :return: The collate function for the PyTorch dataloader which is loading this dataset
        :rtype: Callable[[List[Any]], Input]
        """

        # Need to determine the collate function details for each of the samples
        feature_collate_fns = {}
        for elem in train_data_config.features + val_data_config.features:
            if elem['output_key'] not in feature_collate_fns:
                feature_collate_fns[elem['output_key']] = elem['type'].get_collate_fn(
                    **{k: v for k, v in elem.items() if k != 'type'})

        def _collate_fn(batch):
            output_element: Input = {}
            for key, cfn in feature_collate_fns.items():
                output_element.update(cfn([b[key] for b in batch]))
            output_element.update({'num_samples': len(batch)})
            return output_element

        return _collate_fn
