from typing import List, Dict, cast

import torch
import yaml
try:
    from yaml import CLoader as Loader, CDumper as Dumper  # noqa
except ImportError:
    from yaml import Loader, Dumper  # noqa
    print('Falling back to Non-C-loader')

import numpy as np

from pele.core.data.features import DataFeature
from pele.nn.functional.util import padded_stack


class NumpyFeature(DataFeature):

    def __init__(self, yaml_path: str, output_key: str, pad: bool = False) -> None:
        self._yaml_path = yaml_path
        self._output_key = output_key
        self._pad = pad

        # Load the YAML file
        with open(self._yaml_path, 'r') as yaml_file:
            self._data = yaml.load(yaml_file, Loader=Loader)

        # Save the feature metadata
        self._shape = self._data['metadata']['shape']
        self._samples = self._data['samples']

    def keys(self,) -> List[str]:
        return list(self._samples.keys())

    def shape(self,) -> List[int]:
        return cast(List[int], self._shape)

    def __getitem__(self, key):
        return {self._output_key: np.load(self._samples[key])}

    @classmethod
    def get_collate_fn(cls, output_key: str, pad: bool = False, **kwargs):  # pylint:disable=unused-argument

        def _collate(batch: List[np.ndarray]) -> Dict[str, torch.Tensor]:
            if pad:
                return {
                    output_key: padded_stack([torch.tensor(b) for b in batch]),  # pylint:disable=not-callable
                    output_key + '_seqlen': torch.tensor([len(b) for b in batch])  # pylint:disable=not-callable
                }
            return {output_key: torch.stack([torch.tensor(b) for b in batch], dim=0)}  # pylint:disable=not-callable

        return _collate
