from typing import Optional
import logging

import torch
from torch.optim import Optimizer  # noqa

from ray.util.sgd.torch.training_operator import TrainingOperator

try:
    from apex import amp
except ImportError:
    # Apex library is not installed, so we cannot enable mixed precision.
    # We don't log here because logging happens in the torch_runner,
    # where amp is initialized.
    amp = None

# These wildcard imports are required for registering the gin configurables
# and it's a lot cleaner to do it this way than by trying to figure out the details
from pele.core.models import *  # noqa
from pele.core.fusion import *  # noqa
from pele.core.losses import *  # noqa
from pele.core.data.features import *  # noqa

from pele.core.model import ExperimentModel
from pele.core.loss import ExperimentLoss
from pele.core.data import ExperimentDataset
from pele.config import (ApexConfig, DataConfig, FeatureExtractorConfig, FusionConfig, LossConfig, ModelConfig,
                         TrainConfig, LogConfig, RayConfig, TuneConfig)


class _ExperimentTrainingOp(TrainingOperator):

    def train_batch(self, batch, batch_info):
        if torch.cuda.is_available():
            # If it's not already a CUDA tensor, we should move the input tensors to cuda
            batch = {k: v.cuda(non_blocking=True) if isinstance(
                v, torch.Tensor) else v for k, v in batch.items()}

        # Compute output.
        with self.timers["fwd"]:
            output = self.model(batch)
            metrics = self.criterion(output)
            # torchviz.make_dot(metrics['loss'], params=dict(self.model.named_parameters())).render('./graph.gv')
            print('(Rank {}) Global Step: {}, Train Loss: {}'.format(
                torch.distributed.get_rank() if torch.distributed.is_initialized(
                ) else 0, batch_info['global_step'],
                metrics['loss'].item()))

        # Compute gradients in a backward pass.
        with self.timers["grad"]:
            if self.use_fp16:
                if batch_info['global_step'] % self.criterion.gradient_acc_steps == 0:
                    # Time for a gradient pass
                    with amp.scale_loss(metrics['loss'], self.optimizer) as scaled_loss:
                        scaled_loss.backward()
                    with self.timers["apply"]:
                        torch.nn.utils.clip_grad_norm_(amp.master_params(self.optimizer),
                                                       self.criterion.max_gradient_norm)
                        self.optimizer.step()
                    self.optimizer.zero_grad()
                else:
                    with amp.scale_loss(metrics['loss'], self.optimizer, delay_unscale=True) as scaled_loss:
                        scaled_loss.backward()
            else:
                metrics['loss'].backward()
                if batch_info['global_step'] % self.criterion.gradient_acc_steps == 0:
                    with self.timers["apply"]:
                        torch.nn.utils.clip_grad_norm_(amp.master_params(self.optimizer),
                                                       self.criterion.max_gradient_norm)
                        self.optimizer.step()
                    self.optimizer.zero_grad()

        outputs = {"loss": metrics['loss'].item(
        ), "num_samples": batch['num_samples']}
        outputs = outputs.update(
            metrics['reported']) if 'reported' in metrics else outputs

        if self.criterion.use_wandb and torch.distributed.get_rank() <= 0:
            import wandb
            wandb.log({'train_'+k: v for k, v in outputs.items()})

        return outputs

    def validate_batch(self, batch, batch_info):  # pylint:disable=unused-argument
        if torch.cuda.is_available():
            # If it's not already a CUDA tensor, we should move the input tensors to cuda
            batch = {k: v.cuda(non_blocking=True) if isinstance(
                v, torch.Tensor) else v for k, v in batch.items()}

        # Compute output.
        with self.timers["fwd"]:
            output = self.model(batch)
            metrics = self.criterion(output)
            print(
                'Eval -- Step: {}, Loss: {}'.format(batch_info['batch_idx'], metrics['loss'].item()))

        outputs = {"loss": metrics['loss'].item(
        ), "num_samples": batch['num_samples']}
        outputs = outputs.update(
            metrics['reported']) if 'reported' in metrics else outputs

        if self.criterion.use_wandb and torch.distributed.get_rank() <= 0:
            import wandb
            wandb.log({'val_'+k: v for k, v in outputs.items()})

        return outputs


class Experiment():
    """
    This is the core experiment class. It allows for building and baking each of the
    different model functions which we use while training.
    """

    def __init__(self,
                 train_config: TrainConfig,
                 model_config: ModelConfig,
                 train_data_config: DataConfig,
                 val_data_config: DataConfig,
                 loss_config: LossConfig,
                 fe_config: Optional[FeatureExtractorConfig] = None,
                 fusion_config: Optional[FusionConfig] = None,
                 apex_config: Optional[ApexConfig] = None,
                 logging_config: Optional[LogConfig] = None,
                 ray_config: Optional[RayConfig] = None,
                 tune_config: Optional[TuneConfig] = None,
                 ):   # pylint:disable=too-many-arguments

        self.train_config = train_config
        self.model_config = model_config
        self.train_data_config = train_data_config
        self.val_data_config = val_data_config
        self.fe_config = fe_config
        self.fusion_config = fusion_config
        self.loss_config = loss_config
        self.apex_config = apex_config
        self.logging_config = logging_config
        self.ray_config = ray_config
        self.tune_config = tune_config

    def model_creator(self, config):
        # Return the model function

        # Update the model config with any model information passed in the ray
        # configuration
        if 'model' in config:
            self.model_config.args.update(config['model'])

        if 'tunables' in config:
            self.model_config.replace_tunables(config['tunables'])
            self.train_config.replace_tunables(config['tunables'])
            self.fusion_config.replace_tunables(config['tunables'])
            self.fe_config.replace_tunables(config['tunables'])
            self.logging_config.replace_tunables(config['tunables'])

        logging.info('Building model...')
        model = ExperimentModel(model_config=self.model_config,
                                train_config=self.train_config,
                                fusion_config=self.fusion_config,
                                feature_extraction_config=self.fe_config)

        if self.logging_config.use_wandb and torch.distributed.get_rank() <= 0:
            import wandb
            wandb.watch(model)

        return model

    def data_creator(self, config):  # pylint:disable=unused-argument

        if 'tunables' in config:
            self.train_data_config.replace_tunables(config['tunables'])
            self.val_data_config.replace_tunables(config['tunables'])

        # Return the training and validation datasets
        logging.info('Constructing training dataset...')
        train_dataset = ExperimentDataset(self.train_data_config)
        logging.info('Constructing validation dataset')
        val_dataset = ExperimentDataset(self.val_data_config)

        return train_dataset, val_dataset

    def optimizer_creator(self, model, config):
        # Update the optimizer config with anything from the config variables
        if 'optimizer' in config:
            self.train_config.optimizer_args.update(config['optimizer'])

        if 'tunables' in config:
            self.train_config.replace_tunables(config['tunables'])

        logging.info('Building optimizer...')
        return self.train_config.optimizer(params=model.parameters(), **self.train_config.get_optimizer_kwargs())

    def loss_creator(self, config):  # pylint:disable=unused-argument
        # Return the loss function which will be used on the data
        logging.info('Building loss functions...')

        if 'tunables' in config:
            self.loss_config.replace_tunables(config['tunables'])
            self.train_config.replace_tunables(config['tunables'])
            self.logging_config.replace_tunables(config['tunables'])

        loss_module = ExperimentLoss(
            loss_config=self.loss_config,
            train_config=self.train_config,
            logging_config=self.logging_config)

        return loss_module

    def scheduler_creator(self, optimizer, config):
        # Update the scheduler config with anything from the config variables
        if 'scheduler' in config:
            self.train_config.lr_schedule_args.update(config['scheduler'])

        if 'tunables' in config:
            self.train_config.replace_tunables(config['tunables'])

        logging.info('Building scheduler...')
        return self.train_config.lr_schedule(optimizer, **self.train_config.get_lr_schedule_kwargs())

    def get_training_operator_cls(self,):  # pylint:disable=no-self-use
        # Build and return the training operator class which will be used
        return _ExperimentTrainingOp

    def get_dataloader_config(self, ray_config):
        logging.info('Constructing collate function...')
        return {
            'collate_fn': ExperimentDataset.build_collate_fn(self.train_data_config, self.val_data_config),
            'num_workers': ray_config.num_cpus_per_replica
        }
