import gin

import torch

from pele.types import Input
from pele.nn.functional.masking import sequence_mask


@gin.configurable('LMLoss')
class LanguageModelingLoss(torch.nn.Module):

    def __init__(self, input_key=gin.REQUIRED, target_key=gin.REQUIRED, seqlen_key=gin.REQUIRED, loss_weight=1.0):
        super(LanguageModelingLoss, self).__init__()
        self.loss_weight = loss_weight
        self.input_key = input_key
        self.target_key = target_key
        self.seqlen_key = seqlen_key

    def forward(self, inputs: Input) -> torch.Tensor:  # pylint,disable=arguments-differ
        # Make sure that we have the right keys in the input
        assert self.target_key in inputs
        assert self.input_key in inputs
        assert self.seqlen_key in inputs

        gt = inputs[self.target_key]
        predictions = inputs[self.input_key]
        seqlen = inputs[self.seqlen_key]

        # Shift the ground truth and the predictions
        targets = gt[:, 1:]  # Need to shift one to the left
        predictions = predictions[:, :-1]  # Since GT shifted 1 left, cut last from logits

        # Compute the mask weights for the loss
        mask_weights = sequence_mask(targets, seqlen - 1).type_as(predictions)

        # Reshape the outputs
        predictions = predictions.reshape(-1, predictions.shape[-1])
        targets = targets.reshape(-1)
        mask_weights = mask_weights.reshape(-1)

        # Compute the per-word losses
        loss = torch.nn.functional.cross_entropy(predictions, targets, reduction='none')
        masked_loss = loss * mask_weights
        final_loss = (masked_loss.sum() / seqlen.sum()) * self.loss_weight  # Re-weight the loss

        return final_loss
