from contextlib import nullcontext
from typing import Callable, Optional, cast

import torch

from pele.types import Input
from pele.config import FeatureExtractorConfig, FusionConfig, ModelConfig, TrainConfig


class ExperimentModel(torch.nn.Module):

    def __init__(
            self,
            model_config: ModelConfig,
            train_config: TrainConfig,
            fusion_config: Optional[FusionConfig] = None,
            feature_extraction_config: Optional[FeatureExtractorConfig] = None,
    ) -> None:

        super(ExperimentModel, self).__init__()

        # Construct the individual feature extractors
        self.feature_extractors = torch.nn.ModuleList()
        self.fe_compute_gradients = []
        if feature_extraction_config:
            for input_key, extractor_chain in feature_extraction_config.feature_extractor_map.items():
                for model in extractor_chain:
                    self.feature_extractors.append(
                        model['type'](**{k: v for k, v in model.items() if k not in ('type', 'compute_gradients')}))
                    self.fe_compute_gradients.append(model['compute_gradients'] if 'compute_gradients' in
                                                     model else False)

        # Construct the fusion model
        self.fusion_model: Optional[torch.nn.Module] = fusion_config.model['type'](
            **{k: v for k, v in fusion_config.model.items() if k != 'type'}) if fusion_config else None

        # Construct the downstream model
        self.downstream_model = model_config.model['type'](**{k: v for k, v in model_config.model.items() if k != 'type'})

        # Setup some training details
        self.use_teacher_forcing: bool = train_config.use_teacher_forcing
        self.stopping_criterion: Callable[[Input], bool] = lambda x: False
        if self.use_teacher_forcing is not True:
            raise ValueError('[TODO] - Autoregression is not yet implemented, set use_teacher_forcing to False')

    def forward(self, inputs: Input) -> Input:

        # Step 1: Extract features from the inputs
        for idx, feature_extractor in enumerate(self.feature_extractors):
            with torch.no_grad() if not self.fe_compute_gradients[idx] else nullcontext():
                inputs.update(feature_extractor(inputs=inputs))

        # Step 2: Run the fusion model to get the context input
        if self.fusion_model:
            self.fusion_model = cast(torch.nn.Module, self.fusion_model)
            inputs.update(self.fusion_model(inputs=inputs))

        # Step 3: Run the downstream model to get the outputs
        if self.use_teacher_forcing:
            inputs.update(self.downstream_model(inputs=inputs))
        else:
            while self.stopping_criterion(inputs):
                inputs.update(self.downstream_model(inputs=inputs))

        return inputs
