import torch
import gin


@gin.configurable('LSTMDescriber')
class LSTMDescriber(torch.nn.Module):

    def __init__(self, internal_dimension, context_dimension, vocab_size, num_layers=2, dropout=0.1):
        super(LSTMDescriber, self).__init__()

        self.hidden_dimension = internal_dimension
        self.context_dimension = context_dimension
        self.vocab_size = vocab_size
        self.num_layers = num_layers
        self.dropout = dropout

        # Model
        self.video_encoder = torch.nn.LSTM(self.hidden_dimension,
                                           self.hidden_dimension,
                                           batch_first=True,
                                           dropout=self.dropout)
        self.text_encoder = torch.nn.LSTM(self.hidden_dimension,
                                          self.hidden_dimension,
                                          batch_first=True,
                                          dropout=self.dropout)
        self.output_decoder = torch.nn.LSTM(2 * self.hidden_dimension,
                                            self.hidden_dimension,
                                            num_layers=self.num_layers,
                                            batch_first=True,
                                            dropout=self.dropout)

        # Projection Layer
        self.projection_layer = torch.nn.Linear(self.hidden_dimension, self.vocab_size)
        self.context_projection_layer = torch.nn.Linear(self.context_dimension, self.hidden_dimension)

    def forward(self, inputs):
        # Initial states
        features_encoded, _ = self.video_encoder(
            torch.nn.functional.relu(
                self.context_projection_layer(inputs['context'])))
        caption_encoded, _ = self.text_encoder(inputs['caption_embedded'])

        input_tensors = torch.cat([
            torch.nn.functional.relu(features_encoded.mean(dim=1).unsqueeze(1).expand(-1, caption_encoded.shape[1], -1)),
            torch.nn.functional.relu(caption_encoded),
        ], dim=-1)

        # Output Decoder
        decoded, _ = self.output_decoder(input_tensors)
        text_logits = self.projection_layer(decoded)

        return {
            'predicted_caption_logits': text_logits,
        }
