import logging
import collections
from typing import Dict


import click
import gin
import ray
from ray import tune as ray_tune
from ray.util.sgd.torch.torch_trainer import TorchTrainable
import torch

from pele.run import main_fn
from pele.core import Experiment
from pele.logging import init_logging
from pele.config import ApexConfig, RayConfig, TrainConfig, LogConfig, DataConfig, ModelConfig, FusionConfig, FeatureExtractorConfig, LossConfig, TuneConfig
from pele.config.util import ConfigClass, TUNABLES


@main_fn.command(context_settings={
    'ignore_unknown_options': True,
    'allow_extra_args': True,
})
@click.pass_context
@click.argument('config-file',
                default=None,
                type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True))
def tune(context, config_file):
    if config_file is None:
        raise RuntimeError('You must specify a run configuration')

    # Parse the gin configuration for the experiment
    gin.parse_config_file(config_file)
    gin.parse_config([arg[2:] for arg in context.args])  # Parse overrides from the command line

    # Setup the configurations
    configs: Dict[ConfigClass] = {
        'apex_config': ApexConfig(),
        'ray_config': RayConfig(),
        'train_config': TrainConfig(),
        'logging_config': LogConfig(),
        'model_config': ModelConfig(),
        'fusion_config': FusionConfig(),
        'fe_config': FeatureExtractorConfig(),
        'loss_config': LossConfig(),
        'tune_config': TuneConfig(),
    }

    with gin.config_scope('train'):
        configs.update({'train_data_config': DataConfig()})
    with gin.config_scope('val'):
        configs.update({'val_data_config': DataConfig()})

    # Initialize ray if we haven't already specified a tuning job
    ray.init(num_cpus=configs['ray_config'].num_replicas * configs['ray_config'].num_cpus_per_replica)

    # Setup the logging
    init_logging(configs['logging_config'], rank=torch.distributed.get_rank())

    # Compute the dictionary configuration
    for k, v in configs.items():
        v.process_tunables(k)

    logging.info('Tuning %i features: %s', len(TUNABLES), str(TUNABLES.keys()))

    # Build the experiment and visualization
    experiment = Experiment(
        **configs,  # Pass the configs dictionary to the experiment constructor as kwargs
    )

    config = {
        'model_creator': experiment.model_creator,
        'data_creator': experiment.data_creator,
        'dataloader_config': experiment.get_dataloader_config(configs['ray_config']),
        'optimizer_creator': experiment.optimizer_creator,
        'loss_creator': experiment.loss_creator,
        'scheduler_creator': experiment.scheduler_creator,
        'training_operator_cls': experiment.get_training_operator_cls(),
        'num_replicas': configs['ray_config'].num_replicas,
        'use_gpu': configs['ray_config'].use_gpu,
        'use_fp16': configs['ray_config'].use_fp16,
        'apex_args': {
            'opt_level': configs['apex_config'].opt_level,
            'num_losses': configs['loss_config'].num_losses,
            'verbosity': configs['apex_config'].verbosity
        },
        'batch_size': configs['train_config'].batch_size,
        'config': {
            'tunables': TUNABLES
        },
    }

    metric_columns = collections.OrderedDict({
        'mean_validation_loss': "validation loss",
        ray_tune.result.TIME_TOTAL_S: "total time (s)",
        ray_tune.result.TIMESTEPS_TOTAL: "ts",
    })

    analysis = ray_tune.run(
        TorchTrainable,
        num_samples=configs['tune_config'].num_samples,
        config=config,
        stop={'epoch': 7},
        verbose=1,
        progress_reporter=ray_tune.CLIReporter(metric_columns=metric_columns)
    )
    print(analysis.get_best_config(metric='mean_validation_loss', mode='min').get('config'))
